import rasterio
from rasterio.vrt import WarpedVRT


def warp(img, transform, source_crs, crs, resampling):
    """
    Warp an image.
    """
    band, height, width = img.shape
    with rasterio.MemoryFile() as memfile:
        with memfile.open(
                driver="GTiff",
                height=height,
                width=width,
                count=band,
                dtype=str(img.dtype.name),
                crs=source_crs,
                transform=transform,
        ) as dataset:
            dataset.write(img)

        with memfile.open() as src:
            with WarpedVRT(src, crs=crs, resampling=resampling) as vrt:
                ds = vrt.read()
                bounds = vrt.bounds
                transform = vrt.transform

    return ds, bounds, transform



import logging
import os
import typing
from dataclasses import dataclass

from movdata import kde_range, raster, base

logger = logging.getLogger(__name__)


@dataclass
class KDERange:
    multirelocation: base.MultiRelocations
    output_dir: typing.Union[str, bytes, os.PathLike]
    pixel_size: float = 250.
    crs: typing.Any = 'EPSG:8857'
    nodata_value: float = 0.
    smooth_param: float = 0.0
    max_sd: float = 3.0
    prob_dens_cut_off: float = 1E-15
    expansion_factor: float = 1.

    def __post_init__(self):
        # assert type(self.multirelocation) == base.Relocations
        self.raster_profile = raster.RasterProfile(crs=self.crs,
                                                   pixel_size=self.pixel_size,
                                                   nodata_value=self.nodata_value
                                                   )
        self.output_raster = {}

        def _calculate_kde_range_(df):
            subject_name = df.subject_id.unique()[0]
            logger.info(f"Calculating KDE Range for subject {subject_name}")

            # replace whitespace with underscore char (_)
            subject_name = subject_name.replace(" ", "_")

            # calculate KDE range
            kde_profile = kde_range.KDEAnalysisProfile(smooth_param=self.smooth_param,
                                                       max_sd=self.max_sd,
                                                       prob_dens_cut_off=self.prob_dens_cut_off,
                                                       raster_profile=self.raster_profile,
                                                       expansion_factor=self.expansion_factor,
                                                       output_path=os.path.join(self.output_dir, f'{subject_name}.tif'))
            kde_range.KDEAnalysis.calculate_kde_range(kde_profile, df)
            self.output_raster[str(subject_name)] = os.path.join(self.output_dir, f'{subject_name}.tif')

        self.multirelocation.df().groupby('subject_id').apply(_calculate_kde_range_)



import os
import typing
from movdata.percentile_area import PercentileArea, PercentileAreaProfile


def get_percentile_area(percentile_levels: typing.List,
                        raster_path: typing.Union[str, bytes, os.PathLike],
                        subject_id: str = ''):
    percentile_profile = PercentileAreaProfile(percentile_levels=percentile_levels,
                                               input_raster=raster_path,
                                               subject_id=subject_id)
    return PercentileArea.calculate_percentile_area(profile=percentile_profile)



import datetime
import pandas
import numpy
import ee


def extract_ee_img_coll_vals(region=None, img_coll_name='MODIS/006/MOD13Q1',
                             band_name='NDVI', scaling_const=0.0001, start=None,  end=None,
                             raster_resolution=500.0):

    ee.Initialize()

    def convertUnixTimeToDateTime(unixTime):
        t = datetime.datetime.utcfromtimestamp(int(unixTime / 1000))
        return t.strftime('%Y-%m-%d %H:%M:%S')

    def convertDateTimeToUnixTime(inputTime):
        epoch = datetime.datetime(1970, 1, 1, 0, 0, 0)
        return int((inputTime - epoch).total_seconds() * 1000)

    def ComputePolygonTimeSeries(region, t1, t2):

        """Returns a series of brightness over time for the polygon."""
        #     d2=datetime.strptime(t2_String,'%Y-%m-%d %H:%M:%S')
        #     d1=datetime.strptime(t1_String,'%Y-%m-%d %H:%M:%S')

        collection_now = ee.ImageCollection(img_coll_name).filterDate(convertDateTimeToUnixTime(t1),
                                                                      convertDateTimeToUnixTime(t2))

        collection_historic = ee.ImageCollection(img_coll_name).filter(ee.Filter.dayOfYear(t1.timetuple().tm_yday,
                                                                                           t2.timetuple().tm_yday))
        collection_now = collection_now.sort('system:time_start')
        collection_historic = collection_historic.sort('system:time_start')

        # Compute the mean brightness in the region in each image.
        def ComputeMean(img):
            reduction = img.select(band_name).reduceRegion(ee.Reducer.mean(),
                                                           region,
                                                           raster_resolution,
                                                           bestEffort=True)

            return ee.Feature(None, {band_name: reduction.get(band_name),
                                     'system:time_end': img.get('system:time_end')})

        # Extract the results as a list of lists.
        def ExtractMean(feature):
            return (feature.get('properties').get('system:time_end'),
                    feature.get('properties').get(band_name)) or numpy.NaN

        cur_data = collection_now.map(ComputeMean).getInfo()
        cur_data = map(ExtractMean, cur_data['features'])
        cur_vals = [(convertUnixTimeToDateTime(x[0]), x[1]) for x in cur_data]
        cur_data = pandas.DataFrame(cur_vals, columns=['time', band_name])
        cur_data[band_name] = cur_data[band_name] * scaling_const
        cur_data[band_name] = pandas.to_numeric(cur_data[band_name])
        cur_data.dropna(inplace=True)

        hist_data = collection_historic.map(ComputeMean).getInfo()
        hist_data = map(ExtractMean, hist_data['features'])
        hist_vals = [(convertUnixTimeToDateTime(x[0]), x[1]) for x in hist_data]
        hist_data = pandas.DataFrame(hist_vals, columns=['time', band_name])
        hist_data[band_name] = hist_data[band_name] * scaling_const
        hist_data[band_name] = pandas.to_numeric(hist_data[band_name])
        hist_data.dropna(inplace=True)

        return cur_data, hist_data

    cur_data, hist_data = ComputePolygonTimeSeries(region, t1=start, t2=end)

    cur_data['time'] = pandas.to_datetime(cur_data['time'])
    cur_data['day_of_year'] = cur_data['time'].apply(lambda x: int(x.timetuple().tm_yday))

    hist_data['time'] = pandas.to_datetime(hist_data['time'])
    hist_data['day_of_year'] = hist_data['time'].apply(lambda x: int(x.timetuple().tm_yday))

    # Calculate the mean & std dev of the historic values
    df_mean = hist_data.groupby(['day_of_year'])[band_name].agg([numpy.mean, numpy.std])
    df_mean.head()

    # Merge historic avg/stddev data with cur values
    final_df = pandas.merge(df_mean, cur_data, left_index=True, right_on='day_of_year', how='inner')
    final_df.sort_values('time', inplace=True)
    final_df.set_index('time', drop=True, inplace=True)
    return final_df


import geopandas as gpd
import leafmap
import leafmap.foliumap
import mapclassify
import pandas as pd
from shapely.geometry import MultiLineString
import typing

ESRI_TOPOLOGY_BASEMAP = 'https://services.arcgisonline.com/ArcGIS/rest/services/World_Topo_Map/MapServer/tile/{z}/{y}/{x}'


class EcoMapMixin():
    speedmap_labels = []

    @property
    def default_speed_color(self):
        return ['#1a9850', '#91cf60', '#d9ef8b', '#fee08b', '#fc8d59', '#d73027']

    @staticmethod
    def _speedmap_labels(bins):
        return [f'{bins[i]:.1f} - {bins[i + 1]:.1f} km/hr' for i in range(len(bins) - 1)]

    @property
    def classification_methods(self):
        return {
            "equal_interval": mapclassify.EqualInterval,
            "natural_breaks": mapclassify.NaturalBreaks,
            "quantile": mapclassify.Quantiles,
            "std_mean": mapclassify.StdMean,
            "max_breaks": mapclassify.MaximumBreaks,
            "fisher_jenks": mapclassify.FisherJenks,
        }

    @staticmethod
    def create_multi_linestring(s):
        return MultiLineString(s['geometry'].tolist())

    def apply_classification(self, x, k, cls_method='natural_breaks', multiples=None):
        """
        Function to select which classifier to apply to the speed distributed data.

        Args:
        __________
        x (array)          : The input array to be classified. Must be 1-dimensional
        k (int)            : Number of classes required.
        cls_method (str)   : Classification method
        multiples (array)  : the multiples of the standard deviation to add/subtract from
                             the sample mean to define the bins.
                             defaults=[-2,-1,1,2]
        """
        if multiples is None:
            multiples = [-2, -1, 1, 2]

        classifier = self.classification_methods.get(cls_method)
        if not classifier:
            return

        map_classifier = classifier(x, multiples) if cls_method == 'std_mean' else classifier(x, k)
        edges, _, _ = mapclassify.classifiers._format_intervals(map_classifier, fmt="{:.2f}")

        return [float(i) for i in edges]

    def create_speedmap_df(
        self,
        multi_trajectory: gpd.GeoDataFrame,
        classification_method: str = 'equal_interval',
        no_class: int = 6,
        speed_colors: typing.List = None,
        bins: typing.List = None,
    ):
        if not bins:
            # apply classification on speed data.
            bins = self.apply_classification(multi_trajectory.speed_kmhr, no_class, cls_method=classification_method)
        else:
            no_class = len(bins) - 1

        if speed_colors is None:
            speed_colors = self.default_speed_color[:no_class]

        multi_trajectory['speed_colour'] = pd.cut(x=multi_trajectory.speed_kmhr,
                                                  bins=bins,
                                                  labels=speed_colors,
                                                  include_lowest=True)
        # Group the data according to speed and create multi-linestrings for each group
        speedmap_df = gpd.GeoDataFrame(geometry=multi_trajectory.groupby('speed_colour').apply(
            self.create_multi_linestring),
                                       crs=4326).reset_index()

        speedmap_df.sort_values(by='speed_colour', inplace=True)
        self.speedmap_labels = self._speedmap_labels(bins)
        return speedmap_df

    def add_speedmap(self,
                     multi_trajectory: gpd.GeoDataFrame,
                     basemap=ESRI_TOPOLOGY_BASEMAP,
                     classification_method: str = 'equal_interval',
                     linewidth: float = 1.,
                     no_class: int = 6,
                     speed_colors: typing.List = None,
                     bins: typing.List = None,
                     **kwargs: typing.Any):

        df = self.create_speedmap_df(multi_trajectory=multi_trajectory,
                                     classification_method=classification_method,
                                     no_class=no_class,
                                     speed_colors=speed_colors,
                                     bins=bins)

        self.add_gdf(df, style_callback=lambda x: dict(color=x['properties']['speed_colour']))

        df['speed_colour'] = df['speed_colour'].apply(lambda x: x[1:])
        self.add_legend(legend_dict=dict(zip(self.speedmap_labels, df.speed_colour)))


class EcoMapFolium(EcoMapMixin, leafmap.foliumap.Map):
    pass


EcoMap = EcoMapFolium


class EcoMapIPyLeaflet(EcoMapMixin, leafmap.Map):
    pass



import multiprocessing
import os
import ast
import typing
from collections import OrderedDict

import geopandas as gp
import pandas as pd
from shapely.geometry import Point
from movdata.base import BaseMixin

import sys
import logging

logging.basicConfig(
    format='%(asctime)s [%(levelname)s] %(name)s - %(message)s',
    level=logging.INFO,
    datefmt='%Y-%m-%d %H:%M:%S',
    stream=sys.stdout,
)

unpack_columns = BaseMixin.unpack_columns
pack_columns = BaseMixin.pack_columns


def geometry_point(lat_lon):
    if lat_lon and not pd.isna(lat_lon):
        x, y = float(lat_lon.get('longitude')), float(lat_lon.get('latitude'))
        return Point(x, y)
    return Point()


def set_index_column(df, col_name):
    index_column = df.pop(col_name)

    df.insert(0, col_name, index_column)
    return df.set_index(col_name)


def tz_naive(d):
    return d.tzinfo is None or d.tzinfo.utcoffset(d) is None


def export(df, filename='ouput.csv', layer=None, mode='w', **kwargs):
    if isinstance(df, pd.DataFrame) and filename.endswith('.csv'):
        df.to_csv(filename)
    # writing to spatial data.
    elif isinstance(df, gp.GeoDataFrame):
        if filename.endswith('.shp'):
            df.to_file(filename, mode=mode)
        elif filename.endswith('.gpkg'):
            df.to_file(filename, driver='GPKG', layer=layer)
    else:
        df.to_file(filename)


def parallelize_df(df: typing.Iterable, func: typing.Callable, processes: int = 10):
    """
    Using pool of worker processes to offload tasks
    # https://docs.python.org/3/library/multiprocessing.html#using-a-pool-of-workers
    """
    processes = processes or os.cpu_count()
    with multiprocessing.pool.Pool(processes) as p:
        df = pd.concat(p.map(func, df))
    df.reset_index(inplace=True, drop=True)
    return df


def normalize_dict(x, colname=None):
    """normalize dict into individual columns"""
    try:
        _x = x.get(colname)
        index = list(x.index)
        if _x:
            _x = OrderedDict(x[colname])
        else:
            _x = OrderedDict()
        for key in _x.keys():
            if key not in index:
                x[str(key)] = _x.get(key)
            else:
                x['_'.join((colname, str(key)))] = _x.get(key)
        x = x.drop(colname)
    except Exception as e:
        print(str(e))
    finally:
        return x


def flatten_dict(x, colname=None):
    """normalize dict into individual columns"""
    try:
        _y = x.get(colname)
        _x = None
        if (_y is not None) and (_y != ''):
            _x = OrderedDict(x[colname])
        else:
            _x = OrderedDict()
        for key in _x.keys():
            x['_'.join((colname, str(key)))] = _x.get(key)
        x = x.drop(colname)
    except Exception as e:
        print(str(e))
    finally:
        return x


def memorize(func):
    cache = {}

    def wrapper(*args):
        if args in cache:
            return cache[args]
        result = func(*args)
        cache[args] = func(*args)
        return result
    return wrapper


def extract_voltage(s):
    additional = ast.literal_eval(s.additional)
    voltage = additional.get('battery', None)  # savannah tracking
    if not voltage:
        voltage = additional.get('mainVoltage', None)  # vectronics
    if not voltage:
        voltage = additional.get('batt', None)  # AWT
    if not voltage:
        voltage = additional.get('power', None)  # Followit
    return voltage



import asyncio
import copy
import datetime
import logging
import math
import pprint
import types

import aiohttp
import backoff
import pandas
from dasclient import dasclient

list_errors = []
logger = logging.getLogger(__name__)


def exception_wrapper(func):
    """A wrapper that handles exception by returning a logger."""

    def inner(*args, **kwargs):
        try:
            return func(*args, **kwargs)
        except dasclient.DasClientNotFound:
            logger.error(f"404 (NotFound)")
        except Exception as exc:
            logger.exception(exc)
            return

    return inner


def backoff_wrapper(func):
    async def wrapper(*args, **kwargs):
        try:
            return await func(*args, **kwargs)
        except aiohttp.ContentTypeError as exc:
            list_errors.append(str(exc))
            return empty_dict()
        except Exception as exc:
            list_errors.append(str(exc))
            return empty_dict()

    return wrapper


def empty_dict():
    yield {}


class EarthRangerClient(dasclient.DasClient):

    def __init__(self,
                 er_server: str,
                 username: str,
                 password: str,
                 provider_key: str = '',
                 client_id: str = 'das_web_client'):
        """initialize the earthranger client"""
        root = '/'.join([er_server, 'api/v1.0'])
        token_url = '/'.join([er_server, 'oauth2/token'])

        super().__init__(username=username,
                         password=password,
                         service_root=root,
                         token_url=token_url,
                         provider_key=provider_key,
                         client_id=client_id)
        pprint.pprint(self.pulse())

    @exception_wrapper
    def _get(self, path, stream=False, **kwargs):
        return super(EarthRangerClient, self)._get(path, stream, **kwargs)

    def get_eventtypes(self, include_inactive=True):
        """Gets the eventtypes"""
        return self._get(f'activity/events/eventtypes?include_inactive={include_inactive}')

    def post_subjectsource(self, subject_id, payload):
        urlpath = f'subject/{subject_id}/sources'
        return self._post(urlpath, payload=payload)


class AsyncEarthRangerClient:
    """Asynchronous EarthRanger Client"""

    def __init__(self, er_client):
        self.client = er_client
        self.service_root = er_client.service_root

    def _earthranger_url(self, path):
        return '/'.join((self.service_root, path))

    @staticmethod
    def _yield_results(results):
        for r in results:
            yield r

    @staticmethod
    def _paginate(params, page):
        p = copy.deepcopy(params)
        p['page'] = page
        return p

    @staticmethod
    def _page_counts(result):
        if isinstance(result, types.GeneratorType):
            logger.error("PageCountError: Unable to get page count.")
            return 0
        else:
            return int(result.get('count'))

    @backoff_wrapper
    @backoff.on_exception(backoff.expo, aiohttp.ClientError, max_tries=10)
    async def _get_method(self, session, urlpath, params=None, with_count=False):
        async with session.get(urlpath, params=params) as response:
            json_data = await response.json()
            return self._yield_results(json_data['data']['results']) if not with_count else json_data['data']

    async def _future_objects(self, session, urlpath, params, page_size):

        rs = await self._get_method(session, urlpath, self._paginate(params, 1), with_count=True)
        rs_count = self._page_counts(rs)
        total_pages = math.ceil(rs_count / page_size)

        params['page_size'] = page_size
        return [asyncio.ensure_future(
            self._get_method(session,
                             urlpath,
                             self._paginate(params, page))) for page in range(1, total_pages + 1)]

    async def get_subject_observations(self,
                                       subject_id,
                                       start=None,
                                       end=None,
                                       filter_flag=0,
                                       include_details=True,
                                       page_size=4000,
                                       count=None,
                                       limit=0,
                                       ):

        connector = aiohttp.TCPConnector(limit=limit)
        async with aiohttp.ClientSession(headers=self.client.auth_headers(), connector=connector) as session:
            params = {}
            page_size = min(page_size, 4000)  # Max page_size supported is 4000
            urlpath = self._earthranger_url('observations')

            if start is not None and isinstance(start, datetime.datetime):
                params['since'] = start.isoformat()
            if end is not None and isinstance(end, datetime.datetime):
                params['until'] = end.isoformat()
            if subject_id:
                params['subject_id'] = subject_id
            else:
                raise ValueError('subject_id or source_id missing')

            params['filter'] = 'null' if filter_flag is None else filter_flag
            params['include_details'] = str(include_details)
            params['page_size'] = 1

            rs = await self._get_method(session, urlpath, self._paginate(params, 1), with_count=True)
            rs_count = self._page_counts(rs)
            total_pages = math.ceil(rs_count / page_size)
            count.append(rs_count)

            params['page_size'] = page_size

            tasks = [asyncio.ensure_future(
                self._get_method(session, urlpath, self._paginate(params, page))) for page in range(1, total_pages + 1)]
            return await asyncio.gather(*tasks)

    async def get_subjectsources(self, subjects, sources, page_size):
        async with aiohttp.ClientSession(headers=self.client.auth_headers()) as session:
            params = {}
            page_size = min(page_size, 4000)  # Max page_size supported is 4000
            urlpath = self._earthranger_url('subjectsources')

            params['subjects'] = subjects
            params['sources'] = sources
            params['page_size'] = 1

            tasks = await self._future_objects(session=session, urlpath=urlpath, page_size=page_size, params=params)
            return await asyncio.gather(*tasks)

    async def get_events(self, page_size, **kwargs):
        async with aiohttp.ClientSession(headers=self.client.auth_headers()) as session:
            params = {k: v for k, v in kwargs.items() if
                      k in ['state', 'event_type',
                            'filter', 'include_notes',
                            'include_related_events',
                            'include_files', 'include_details',
                            'include_updates', 'max_results', 'oldest_update_date']}
            page_size = min(page_size, 4000)  # Max page_size supported is 4000
            urlpath = self._earthranger_url('activity/events')

            tasks = await self._future_objects(session=session, urlpath=urlpath, page_size=page_size, params=params)
            return await asyncio.gather(*tasks)

    @staticmethod
    def yield_sources(results):
        yield results['data']

    @backoff_wrapper
    @backoff.on_exception(backoff.expo, aiohttp.ClientError, max_tries=10)
    async def _get2_method(self, session, urlpath, params=None):
        async with session.get(urlpath, params=params) as response:
            json_data = await response.json()
            return self.yield_sources(json_data)

    async def get_sources(self, sources):
        """Temporary solution """
        async with aiohttp.ClientSession(headers=self.client.auth_headers()) as session:
            urlpath = self._earthranger_url('source/{}')
            tasks = [asyncio.ensure_future(self._get2_method(session, urlpath.format(source_id)))
                     for source_id in sources if not pandas.isnull(source_id)]
            return await asyncio.gather(*tasks)

    @backoff_wrapper
    @backoff.on_exception(backoff.expo, aiohttp.ClientError, max_tries=10)
    async def post_data(self, session, urlpath, observation):
        async with session.post(urlpath, json=observation) as response:
            return await response.text()

    async def post_observations(self, payloads):
        async with aiohttp.ClientSession(headers=self.client.auth_headers()) as session:
            urlpath = self._earthranger_url('observations')
            tasks = [asyncio.ensure_future(self.post_data(session, urlpath, o)) for o in payloads]
            return await asyncio.gather(*tasks)




import typing
import datetime
import ee
import pandas
import logging
import numpy
import sklearn.mixture as gauss
from sklearn.preprocessing import LabelEncoder
from scipy.stats import norm
import movdata.eetools as eetools
from dataclasses import dataclass, field

logger = logging.getLogger(__name__)


@dataclass
class Seasons:
    start: datetime.datetime = None
    end: datetime.datetime = None
    image_col: str = 'MODIS/MCD43A4_006_NDVI'
    band: str = 'NDVI'
    analysis_scale_metre: int = 500
    analysis_geo: typing.Any = None
    analysis_date_chunks: int = 3
    calibration_func: typing.Callable = None
    season_num: int = 2
    season_labels: typing.Any = None
    values: float = field(init=False, repr=False, default=None)
    season_cut: float = field(init=False, repr=False, default=None)

    def extract_seasonal_vals(self):
        # TODO: define an identity calibration function if none exists

        secs = self.end.timestamp() - self.start.timestamp()
        chunk_size = secs // self.analysis_date_chunks
        date_slices = []
        cur_timestamp = self.start.timestamp()
        while cur_timestamp < self.end.timestamp():
            date_slices.append(cur_timestamp)
            cur_timestamp += chunk_size
        date_slices.append(self.end.timestamp() + 1)

        result_df = []

        for i in range(len(date_slices) - 1):

            start = datetime.datetime.fromtimestamp(date_slices[i])
            end = datetime.datetime.fromtimestamp(date_slices[i + 1] - 1)
            print('Processing date range: ', start, end)

            # get the image stack
            img_coll = ee.ImageCollection(self.image_col).select(self.band). \
                filterBounds(self.analysis_geo).filter(ee.Filter.date(start, end)). \
                sort('system:time_start', False)

            # add time info
            tmpColl = img_coll.map(eetools.add_img_time).toArray()

            # extract the pixel values falling within the region and for the given dates
            ee_result = tmpColl.reduceRegion(ee.Reducer.toList(), self.analysis_geo,
                                             self.analysis_scale_metre).getInfo()

            vals = pandas.DataFrame([[x[0], eetools.convertUnixTimeToDateTime(x[1])] for y in ee_result.get('array')
                                     for x in y], columns=['val', 'datetime'])

            if vals is not None:
                if self.calibration_func is not None:
                    vals['val'].apply(self.calibration_func)
                result_df.append(vals)

        self.values = pandas.concat(result_df, axis=0).sort_values(by='datetime')
        return self.values

    def seasonal_cutvals(self):
        try:
            if self.values is not None:
                distr = gauss.GaussianMixture(n_components=self.season_num, max_iter=500)
                vals = self.values['val'].values.reshape(-1, 1)
                distr.fit(vals)
                mu_vars = numpy.array(sorted(zip(distr.means_.flatten(),
                                                 distr.covariances_.flatten()), key=lambda x: x[0]))
                print('mu_vars', mu_vars)
                cuts = []
                x = numpy.linspace(0, 1.0, 1000)
                for i in range(1, len(mu_vars)):
                    cuts.append(numpy.max(x[tuple([norm.sf(x, loc=mu_vars[tuple([i - 1, 0])],
                                                           scale=numpy.sqrt(mu_vars[tuple([i - 1, 1])])) >
                                                   norm.cdf(x, loc=mu_vars[tuple([i, 0])],
                                                            scale=numpy.sqrt(mu_vars[tuple([i, 1])]))])]))
                cuts.append(float("inf"))
                cuts.append(float("-inf"))
                self.season_num = sorted(cuts)

            else:
                logger.exception('Could not run EM algorithm on null vals array.')
        finally:
            return self.season_cut

    def seasonal_windows(self, reducer_func=numpy.mean):
        result = None

        try:
            self.extract_seasonal_vals()
            self.seasonal_cutvals()

            if self.season_cut is None:
                logger.exception('Could not determine_seasonal_windows on null cut values array.')

            if self.values is None:
                logger.exception('Could not determine_seasonal_windows on null values array.')

            def g(group):
                group['reduce_val'] = group['val'].agg(reducer_func)
                return group.iloc[0]  # return first row only

            df = self.values.groupby('datetime').apply(g)
            df.drop(labels=['val'], axis=1, inplace=True)

            season_labels = self.season_labels
            if season_labels is None:
                season_labels = ['season_' + str(x) for x in range(len(self.season_cut))]

            df['season'] = pandas.cut(df['reduce_val'], bins=self.season_cut, labels=season_labels)

            enc = LabelEncoder()
            df['season_code'] = enc.fit_transform(df['season'])

            '''Adapted from https://stackoverflow.com/questions/26911851/
            how-to-use-pandas-to-find-consecutive-same-data-in-time-series'''
            df['Unique Season'] = (df['season_code'].diff(1) != 0).astype('int').cumsum()
            df['end'] = df['datetime'].shift(-1)

            # data = OrderedDict([('start', df.groupby('Unique Season')['datetime'].first()),
            #                     ('end', df.groupby('Unique Season')['end'].last()),
            #                     ('season', df.groupby('Unique Season')['season'].first())])
            #
            # result = pd.DataFrame(data, columns=data.keys())

            result = pandas.concat([pandas.Series([grp['datetime'][0] for name, grp in df.groupby('Unique Season')]),
                                    pandas.Series([grp['datetime'][-1] for name, grp in df.groupby('Unique Season')]),
                                    pandas.Series([name for name, _ in df.groupby('Unique Season')]),
                                    pandas.Series(grp['season'][0] for name, grp in df.groupby('Unique Season'))], axis=1)
            result.columns = ['start', 'end', 'unique_season', 'season']
            result.set_index('unique_season', inplace=True)

        except Exception as e:
            print(e)

        finally:
            return result



import matplotlib.ticker
import typing
import ast
import uuid
import logging

import geopandas
import mapclassify
import pandas
import matplotlib.pyplot as plt
import matplotlib.dates as mdates
import numpy as np
import seaborn as sns
import movdata
from shapely.geometry import MultiLineString
from ecoscope.plotting import GEOMAP
from ecoscope.io.utils import extract_voltage

ESRI_TOPOLOGY_BASEMAP = 'https://services.arcgisonline.com/ArcGIS/rest/services/World_Topo_Map/MapServer/tile/{z}/{y}/{x}'
plt.rcParams.update({'figure.max_open_warning': 0})
sns.set_theme()
logger = logging.getLogger(__name__)


class SPEEDMAP(GEOMAP):
    speedmap_labels = []

    @property
    def default_speed_color(self):
        return ['#1a9850', '#91cf60', '#d9ef8b', '#fee08b', '#fc8d59', '#d73027']

    @staticmethod
    def _speedmap_labels(bins):
        return [f'{bins[i]:.1f} - {bins[i + 1]:.1f} km/hr' for i in range(len(bins) - 1)]

    @property
    def classification_methods(self):
        return {
            "equal_interval": mapclassify.EqualInterval,
            "natural_breaks": mapclassify.NaturalBreaks,
            "quantile": mapclassify.Quantiles,
            "std_mean": mapclassify.StdMean,
            "max_breaks": mapclassify.MaximumBreaks,
            "fisher_jenks": mapclassify.FisherJenks,
        }

    @staticmethod
    def create_multi_linestring(s):
        return MultiLineString(s['geometry'].tolist())

    def apply_classification(self, x, k, cls_method='natural_breaks', multiples=None):
        """
        Function to select which classifier to apply to the speed distributed data.

        Args:
        __________
        x (array)          : The input array to be classified. Must be 1-dimensional
        k (int)            : Number of classes required.
        cls_method (str)   : Classification method
        multiples (array)  : the multiples of the standard deviation to add/subtract from
                             the sample mean to define the bins.
                             defaults=[-2,-1,1,2]
        """
        if multiples is None:
            multiples = [-2, -1, 1, 2]

        classifier = self.classification_methods.get(cls_method)
        if not classifier:
            return

        map_classifier = classifier(x, multiples) if cls_method == 'std_mean' else classifier(x, k)
        edges, _, _ = mapclassify.classifiers._format_intervals(map_classifier, fmt="{:.2f}")

        print(map_classifier)
        return [float(i) for i in edges]

    def create_speedmap_df(self,
                           multi_trajectory: geopandas.GeoDataFrame,
                           classification_method: str = 'equal_interval',
                           no_class: int = 6,
                           speed_colors: typing.List = None,
                           bins: typing.List = None,
                           ):
        if not bins:
            # apply classification on speed data.
            bins = self.apply_classification(multi_trajectory.speed_kmhr, no_class, cls_method=classification_method)
        else:
            no_class = len(bins) - 1

        if speed_colors is None:
            speed_colors = self.default_speed_color[:no_class]

        multi_trajectory['speed_colour'] = pandas.cut(x=multi_trajectory.speed_kmhr,
                                                      bins=bins, labels=speed_colors,
                                                      include_lowest=True)
        # Group the data according to speed and create multi-linestrings for each group
        speedmap_df = geopandas.GeoDataFrame(geometry=multi_trajectory.groupby('speed_colour').apply(
            self.create_multi_linestring), crs=4326).reset_index()

        speedmap_df.sort_values(by='speed_colour', inplace=True)
        self.speedmap_labels = self._speedmap_labels(bins)
        return speedmap_df

    def _add_layers(self, df, basemap, linewidth):
        self.add_tiled_layer(source=basemap, crs=df.crs)
        self.add_vector_layer(df,
                              color=df.speed_colour.unique(),
                              column='speed_colour',
                              linewidth=linewidth,
                              linestyle='-', marker=0,
                              label=self.speedmap_labels)

    def plot_speedmap(self,
                      multi_trajectory: geopandas.GeoDataFrame,
                      basemap=ESRI_TOPOLOGY_BASEMAP,
                      classification_method: str = 'equal_interval',
                      linewidth: float = 1.,
                      no_class: int = 6,
                      speed_colors: typing.List = None,
                      bins: typing.List = None,
                      **kwargs: typing.Any
                      ):

        df = self.create_speedmap_df(multi_trajectory=multi_trajectory,
                                     classification_method=classification_method,
                                     no_class=no_class,
                                     speed_colors=speed_colors,
                                     bins=bins)
        self._add_layers(df=df, basemap=basemap, linewidth=linewidth)
        return super(SPEEDMAP, self).plot(**kwargs)


# alias
SpeedMap = SPEEDMAP


def timeseries(data: typing.Union[pandas.DataFrame, geopandas.GeoDataFrame],
               x: typing.Union[str, pandas.Series, geopandas.GeoSeries],
               y: typing.Union[str, typing.List[str], pandas.Series, geopandas.GeoSeries],
               upper: typing.Union[str, pandas.Series, geopandas.GeoSeries] = None,
               lower: typing.Union[str, pandas.Series, geopandas.GeoSeries] = None,
               figsize: typing.Tuple[int, int] = (15, 7),
               xlabel: str = '',
               ylabel: str = '',
               title: str = '',
               lineplot_kwgs: typing.Union[typing.Dict, typing.List[typing.Dict]] = None,
               fill_kwargs: typing.Dict = None,
               xlim: typing.Dict = None,
               ylim: typing.Dict = None,
               locator: matplotlib.ticker.Locator = None,
               formater: matplotlib.ticker.Formatter = None,
               legend: bool = True,
               legend_kwgs: typing.Dict = None
               ):
    if lineplot_kwgs is None:
        lineplot_kwgs = dict(linestyle='solid', marker='o', linewidth=1.)
    if fill_kwargs is None:
        fill_kwargs = dict(alpha=0.1, color='green')
    if legend_kwgs is None:
        legend_kwgs = dict(loc=0)

    fig, ax = plt.subplots(figsize=figsize)
    if isinstance(y, list) and isinstance(lineplot_kwgs, list):
        for yvalue, plot_kwargs in zip(y, lineplot_kwgs):
            sns.lineplot(data=data, x=x, y=yvalue, **plot_kwargs)
    elif isinstance(y, list):
        for yvalue in y:
            sns.lineplot(data=data, x=x, y=yvalue, **lineplot_kwgs)
    else:
        sns.lineplot(data=data, x=x, y=y, **lineplot_kwgs)

    if upper is not None and lower is not None:
        if isinstance(upper, str):
            upper = data[upper]
        if isinstance(lower, str):
            lower = data[lower]
        ax.fill_between(x=x, y1=upper, y2=lower, **fill_kwargs)
    if xlim:
        ax.set_xlim(**xlim)
    if ylim:
        ax.set_ylim(**ylim)
    if locator:
        ax.xaxis.set_major_locator(locator)
    if formater:
        ax.xaxis.set_major_formatter(formater)
    if legend:
        ax.legend(**legend_kwgs)
    ax.set_xlabel(xlabel)
    ax.set_ylabel(ylabel)
    ax.set_title(title)
    ax.grid(True)
    fig.autofmt_xdate()
    return ax, fig


def datacount(data, grouping_column, figsize=(20, 10), title='', xlabel=None, ylabel=None, rotation=90, **countplot):
    sns.set(rc={'figure.figsize': figsize})
    ax = sns.countplot(x=grouping_column, data=data, **countplot)
    plt.xticks(rotation=rotation)
    ax.set_title(title)
    ax.set_xlabel(xlabel)
    ax.set_ylabel(ylabel)
    return ax


def seasonal_window(data,
                    x='start',
                    y='season_code',
                    xlabel='',
                    ylabel='Season (1=wet, 0=dry)',
                    draw_style='steps-post',
                    step='post',
                    color='blue',
                    alpha=0.3):
    ax = sns.lineplot(x=x, y=y, data=data, drawstyle=draw_style)
    l1 = ax.lines[0]
    x1 = l1.get_xydata()[:, 0]
    y1 = l1.get_xydata()[:, 1]
    ax.fill_between(x1, y1, step=step, color=color, alpha=alpha)
    ax.set(xlabel=xlabel, ylabel=ylabel)
    return ax


def ndvi_seasonal_transition(vals,
                             season_cut_values,
                             xlabel='',
                             ylabel='',
                             title='',
                             subplot_kwargs=None,
                             histplot_kwargs=None,
                             axvline_kwargs=None):
    xmin = min(vals) - 0.1 * min(vals)
    xmax = max(vals) + 0.1 * min(vals)
    fig, ax = plt.subplots(**subplot_kwargs)
    ax.set_xlim(xmin, xmax)
    ax.set_title(title)
    ax.set_xlabel(xlabel)
    ax.set_ylabel(ylabel)

    sns.histplot(vals, **histplot_kwargs, ax=ax)
    [ax.axvline(x=i, **axvline_kwargs) for i in season_cut_values[1:-1]]
    return ax


def plot_collar_voltage(multirelocations:movdata.base.MultiRelocations,
                        start_time,
                        extract_fn=extract_voltage,
                        output_folder=None,
                        xhline_kwargs=None):
    multirelocations.unpack_additional(keyname=['subject_name', 'subject_source_id',
                                                'subject_source_end', 'additional'],
                                       colname=['subject_name', 'subject_source_id',
                                                'subject_source_end', 'metadata'])

    multirelocations_df = multirelocations.df().rename(columns={'metadata': 'additional',
                                                                'additional': 'metadata'})
    groups = multirelocations_df.groupby(by=['subject_name', 'subject_source_id'])
    for group, dataframe in groups:
        try:
            subject_source_end = dataframe.subject_source_end.unique()
            is_source_active = subject_source_end >= start_time or pandas.isna(subject_source_end)[0]

            if is_source_active:
                logger.info(group[0])

                dataframe = dataframe.sort_values(by=['fixtime'])
                dataframe['voltage'] = dataframe.apply(extract_fn, axis=1)

                time = dataframe[dataframe.fixtime >= start_time].fixtime.tolist()
                # time_rev = time[::-1]
                voltage = dataframe[dataframe.fixtime >= start_time].voltage.tolist()

                # Calculate the historical voltage
                hist_voltage = dataframe[dataframe.fixtime <= start_time].voltage.tolist()
                if hist_voltage:
                    volt_upper, volt_lower = np.nanpercentile(hist_voltage, [97.5, 2.5])
                    hist_voltage_mean = np.nanmean(hist_voltage)
                else:
                    volt_upper, volt_lower = np.nan, np.nan
                    hist_voltage_mean = None
                volt_diff = volt_upper - volt_lower
                volt_upper = np.full((len(time)), volt_upper, dtype=np.float32)
                volt_lower = np.full((len(time)), volt_lower, dtype=np.float32)

                if np.all(volt_diff == 0):
                    # jitter = np.random.random_sample((len(volt_upper,)))
                    volt_upper = volt_upper + 0.025 * max(volt_upper)
                    volt_lower = volt_lower - 0.025 * max(volt_lower)

                if not any(hist_voltage or voltage):
                    continue

                try:
                    lower_y = min(np.nanmin(np.array(hist_voltage)), np.nanmin(np.array(voltage)))
                    upper_y = max(np.nanmax(np.array(hist_voltage)), np.nanmax(np.array(voltage)))
                except ValueError:
                    lower_y = min(hist_voltage or voltage)
                    upper_y = max(hist_voltage or voltage)
                finally:
                    lower_y = lower_y - 0.1 * lower_y
                    upper_y = upper_y + 0.1 * upper_y

                voltage = np.array(voltage, dtype=np.float32)
                data = pandas.DataFrame({'voltage': voltage, 'time': time,
                                         'volt_lower': volt_lower, 'volt_upper': volt_upper})
                if voltage.size:
                    continue

                # set-ylim
                ylim = {'bottom': lower_y, 'top': upper_y} if not all([np.isnan(lower_y), np.isnan(upper_y)]) else None
                ax, fig = timeseries(data=data,
                                     x=data.time,
                                     y=data.voltage,
                                     figsize=(25, 10),
                                     lower='volt_lower',
                                     upper='volt_upper',
                                     lineplot_kwgs=dict(linestyle='solid', linewidth=1, label=group,
                                                        color='#0000FF'),
                                     fill_kwargs={'color': '#00aeff', 'alpha': 0.2,
                                                  'label': 'Historic 2.5% - 97.5%'},
                                     ylabel='Collar Voltage',
                                     xlabel='Time',
                                     ylim=ylim,
                                     formater=mdates.DateFormatter('%d %b, %y'),
                                     legend=True)
                xhline_kwargs = xhline_kwargs if xhline_kwargs else dict(color='r',
                                                                         linestyle='-',
                                                                         label='Historic average')
                ax.axhline(y=hist_voltage_mean, **xhline_kwargs) if hist_voltage_mean else None
                plt.legend()
                if output_folder:
                    fig.savefig(f'{output_folder}/_{group}_{str(uuid.uuid4())[:4]}.png', dpi=300)
        except ValueError as exc:
            logger.error(exc)



import uuid

import rasterio
from boltons.dictutils import OrderedMultiDict


class LAYERS(OrderedMultiDict):

    def __init__(self, *args, **kwargs):
        super(LAYERS, self).__init__()
        self._clear_ll()

    @property
    def uuid(self):
        return str(uuid.uuid4())

    def tiled_layer(self, source, **plot_kwargs):
        vis_props = dict(source=source, properties=plot_kwargs)
        self.add(f'tiled_layer_{self.uuid}', vis_props)

    def vector_layer(self, geo_dataframe, **plot_kwargs):
        vis_props = dict(df=geo_dataframe, properties=plot_kwargs)
        self.add(f'geo_dataframe_{self.uuid}', vis_props)

    def raster_layer(self, raster_path, resampling=rasterio.enums.Resampling.bilinear, crs=None, **plot_kwargs):
        vis_props = dict(raster=raster_path, resampling=resampling, crs=crs, properties=plot_kwargs)
        self.add(f'raster_layer_{self.uuid}', vis_props)
